/*-----------------------------------------------------------------------

File  : che_heuristics.c

Author: Stephan Schulz

Contents
 
  High-Level interface functions to the heuristics module.


  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jun  8 02:14:51 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_heuristics.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


HeuristicAssocCell HeuristicsTable[]=
{
   {HEU_AUTO_MODE,    "Auto",       HCBAutoModeCreate},
   {HEU_NO_HEURISTIC, NULL,       NULL}
};

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: HCBAutoModeCreate()
//
//   Analyse the proof problem and return an hopefully suitable
//   heuristic. 
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

HCB_p HCBAutoModeCreate(HCBARGUMENTS)
{
   char *res = "Default";
   SpecFeature_p spec = &(control->problem_specs);

   control->selection_strategy = SelectOptimalLiteral;
   OUTPRINT(1, "\n# Auto-Mode is analysing problem.\n");
   
   if(SpecAxiomsAreUnit(spec))
   {
      OUTPRINT(1, "# Axioms are unit-clauses.\n");
      if(SpecGoalsAreGround(spec) || SpecGoalsAreHorn(spec))
      {
	 OUTPRINT(1, "# All goals are ground or some goals are horn.\n");
	 SpecFeaturesAddEval(spec,0.5,0.9,false, AX_SOME_DEFAULT,
			     AX_MANY_DEFAULT, LIT_SOME_DEFAULT,
			     LIT_MANY_DEFAULT, TERM_MED_DEFAULT,
			     TERM_LARGE_DEFAULT );
	 if(SpecGoalsAreHorn(spec))
	 {
	    if(false)
	    {
	       assert(false);
	    }
#include "che_UH____auto.c"     
	 }
	 else
	 {
	    if(false)
	    {
	       assert(false);
	    }
#include "che_UU__G_auto.c"         
	 }
      }
      else
      {
	 OUTPRINT(1, "# Goals are unit and contain variables.\n");
	 SpecFeaturesAddEval(spec,2,5,true, AX_SOME_DEFAULT,
			     AX_MANY_DEFAULT, LIT_SOME_DEFAULT,
			     LIT_MANY_DEFAULT, TERM_MED_DEFAULT,
			     TERM_LARGE_DEFAULT); /* Finalized */
	 if(false)
	 {
	    assert(false);
	 }
#include "che_UU__N_auto.c"         
      }
   }
   else if(SpecAxiomsAreHorn(spec))
   {
      OUTPRINT(1, "# Axioms are horn-clauses.\n");
      SpecFeaturesAddEval(spec, 1,3,true, AX_SOME_DEFAULT,
			  AX_MANY_DEFAULT, LIT_SOME_DEFAULT,
			  LIT_MANY_DEFAULT, TERM_MED_DEFAULT,
			  TERM_LARGE_DEFAULT);
      if(false)
      {
	 assert(false);
      }
#include "che_H_____auto.c" 
   }
   else
   {
      OUTPRINT(1, "# Axioms are general clauses.\n");      
      SpecFeaturesAddEval(spec, FEW_DEFAULT, MANY_DEFAULT, false,
			  AX_SOME_DEFAULT, AX_MANY_DEFAULT,
			  LIT_SOME_DEFAULT, LIT_MANY_DEFAULT,
			  TERM_MED_DEFAULT, TERM_LARGE_DEFAULT);

      if(false)
      {
	 assert(false);
      }
#include "che_G_____auto.c" 
   }
   if(OutputLevel)
   {
      fprintf(stdout, 
	      "# Auto-Mode selected heuristic %s\n"
	      "# and selection function %s.\n", res, 
	      GetLitSelName(control->selection_strategy)); 
   }
   if(parms->mem_limit>2 && (parms->delete_bad_limit ==
			   DEFAULT_DELETE_BAD_LIMIT))
   {
      control->delete_bad_limit =
	 ((float)(parms->mem_limit-2)/64.0)*800000;
      control->filter_copies_limit = control->delete_bad_limit*0.7;
   }
   return GetHeuristic(res, state, control, parms);
}


/*-----------------------------------------------------------------------
//
// Function: HCBCreate() 
//
//   Given a name and the necessary parameters, return a HCB. HCB's
//   must be registered in control->hcbs!
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p HCBCreate(char* name, HCBARGUMENTS)
{
   int    i;
   DStr_p err;

   assert(control->ocb);

   for(i=0; HeuristicsTable[i].heuristic; i++)
   {
      if(strcmp(HeuristicsTable[i].name, name)==0)
      {
	 return HeuristicsTable[i].heuristic_create(state,control,parms);
      }
   }
   err = DStrAlloc();
   DStrAppendStr(err, "Heuristic \"");
   DStrAppendStr(err, name);
   DStrAppendStr(err, "\" unknown");
   Error(DStrView(err), USAGE_ERROR);
   DStrFree(err);
   
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: GetHeuristic()
//
//   Given a string (either a name or a Heuristic-Definition), return
//   a corresponding HCB.
//
// Global Variables: -
//
// Side Effects    : May create HCB's and WFCB's
//
/----------------------------------------------------------------------*/

HCB_p GetHeuristic(char* source, HCBARGUMENTS)
{
   char*     name;
   Scanner_p in;
   HCB_p     res;

   /* StreamTypeOptionString is a hack here...it will give wrong error
      messages for internal strings (but the function will not be
      called with internal strings anyways...) */
   in=CreateScanner(StreamTypeOptionString, source, true, NULL);
   if(TestInpTok(in, OpenBracket))
   {
      HeuristicDefParse(control->hcbs, in, control->wfcbs,
			control->ocb, state);
      name = SecureStrdup("Default");
   }
   else
   {      
      name = SecureStrdup(DStrView(AktToken(in)->literal));
      AcceptInpTok(in, Identifier);
   }
   DestroyScanner(in);
   res = HCBAdminFindHCB(control->hcbs, name);
   if(!res)
   {
      res = HCBCreate(name, state, control, parms);
      assert(res);
   }
   FREE(name);
   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


