/*-----------------------------------------------------------------------

File  : cle_patterns.c

Author: Stephan Schulz

Contents
 

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Wed Apr 14 22:58:47 MET DST 1999
    New

-----------------------------------------------------------------------*/

#include "cle_patterns.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: get_new_fun_symbol()
//
//   Return a new norm-id for a given arity.
//
// Global Variables: -
//
// Side Effects    : Changes used_idents;
//
/----------------------------------------------------------------------*/

static FunCode get_new_fun_symbol(PatternSubst_p subst, int arity)
{
   long    base;
   IntOrP* where;
   
   where = PDArrayElementRef(subst->used_idents, arity);
   where->i_val++;
   base = where->i_val;
   
   assert(base<=NORM_SYMBOL_LIMIT && "Too many function symbols ???");

   return PatternNormCode(base, arity);
}


/*-----------------------------------------------------------------------
//
// Function: extend_pat_subst()
//
//   Extend the pattern subst to substitute symbol (if not already
//   done). Return true if the subst has been extended.
//
// Global Variables: -
//
// Side Effects    : Changes used_idents, extends subst
//
/----------------------------------------------------------------------*/

static bool extend_pat_subst(PatternSubst_p subst, FunCode
				      symbol, int arity)
{
   long res;
   bool result = false;
   
   if(symbol > 0)
   {
      res = PDArrayElementInt(subst->fun_subst, symbol);
      if(!res)
      {
	 res = get_new_fun_symbol(subst, arity);
	 PDArrayAssignInt(subst->fun_subst, symbol, res);
	 PStackPushInt(subst->backtrack, symbol);
	 result = true;
      }      
   }
   else
   {
      assert(symbol < 0);      
      res = PDArrayElementInt(subst->var_subst, -symbol);
      if(!res)
      {
	 res = --subst->used_vars;
	 PDArrayAssignInt(subst->var_subst, -symbol, res);
	 PStackPushInt(subst->backtrack, symbol);
	 result = true;
      }
   }
   return result;
}

/*-----------------------------------------------------------------------
//
// Function: pat_symbol_compare()
//
//   Compare two function symbols as follows:
//
//   If either symbol is unbound but should be bound, return
//    to_uncomparable. Otherwise, compare symbols numerically.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult pat_symbol_compare(PatternSubst_p subst1, FunCode
					f1, PatternSubst_p subst2,
					FunCode f2)
{
   if(PatSymbolIsBound(subst1, f1) && PatSymbolIsBound(subst2, f2))
   {
      FunCode cmp = PatSymbValue(subst1, f1) - PatSymbValue(subst2, f2);
      
      return Q_TO_PART(cmp);
   }
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: pat_term_size_compare()
//
//   Compare two terms with (a variant of) the lexicograpic extension
//   to the ordering induced by the default weights. If two symbols of
//   different arities are encountered, the one with the higher arity
//   is always bigger. This ordering is used as the base for pattern
//   comparison and is independend of actual function symbols. Note
//   that terms with more function symbols are smaller in this
//   ordering!
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static CompareResult pat_term_size_compare(Term_p t1, Term_p t2)
{
   PStack_p      stack = PStackAlloc();
   CompareResult res   = to_equal;
   long          cmp;
   int           i;
   
   PStackPushP(stack, t1);
   PStackPushP(stack, t2);
   
   while(!PStackEmpty(stack))
   {
      t2 = PStackPopP(stack);
      t1 = PStackPopP(stack);
      if(t1==t2)
      {
	 break;
      }
      cmp = TermStandardWeight(t2) - TermStandardWeight(t1);
      
      if(cmp > 0)
      {
	 res = to_greater;
	 break;
      }
      else if(cmp > 0)
      {
	 res = to_lesser;
	 break;
      }
      /* Now either t1 and t2 are variables or neither is! */

      cmp = t1->arity - t2->arity;
      
      if(cmp > 0)
      {
	 res = to_greater;
	 break;
      }
      else if(cmp > 0)
      {
	 res = to_lesser;
	 break;
      }
      for(i=0; i<t1->arity; i++)
      {
	 PStackPushP(stack, t1->args[i]);
	 PStackPushP(stack, t2->args[i]);
      }
   }
   PStackFree(stack);
   assert(res!=to_uncomparable);
   return res;
}

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: PatternSubstAlloc()
//
//   Allocate an empty initialized pattern-substitution cell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

PatternSubst_p PatternSubstAlloc()
{
   PatternSubst_p handle = PatternSubstCellAlloc();

   handle->used_idents = PDIntArrayAlloc(DEFAULT_SIGNATURE_SIZE,
				      DEFAULT_SIGNATURE_SIZE);
   handle->fun_subst   = PDIntArrayAlloc(DEFAULT_SIGNATURE_SIZE,
				      DEFAULT_SIGNATURE_SIZE);
   handle->used_vars   = NORM_VAR_INIT;
   handle->var_subst   = PDIntArrayAlloc(DEFAULT_VARBANK_SIZE,
				      DEFAULT_VARBANK_SIZE);
   handle->backtrack   = PStackAlloc();
   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: PatternDefaultSubsttAlloc()
//
//   Allocate an empty initialized pattern-substitution cell where all
//   system-defined function symbols are bound to themselves.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

PatternSubst_p PatternDefaultSubstAlloc(Sig_p sig)
{
   PatternSubst_p handle = PatternSubstAlloc();
   int i;

   for(i=1; i<=sig->internal_symbols; i++)
   {
      PDArrayAssignInt(handle->fun_subst, i, i);
   }
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PatternSubstFree()
//
//   Free the memory taken by a pattern-subst cell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PatternSubstFree(PatternSubst_p junk)
{
   PDArrayFree(junk->used_idents);
   PDArrayFree(junk->fun_subst);
   PDArrayFree(junk->var_subst);
   PStackFree(junk->backtrack);

   PatternSubstCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PatternSubstCopy()
//
//   Copy a pattern-substitution.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

PatternSubst_p PatternSubstCopy(PatternSubst_p subst)
{
   PatternSubst_p handle = PatternSubstAlloc();
   
   handle->used_idents = PDArrayCopy(subst->used_idents);
   handle->fun_subst   = PDArrayCopy(subst->fun_subst);
   handle->used_vars   = subst->used_vars;
   handle->var_subst   = PDArrayCopy(subst->var_subst);
   handle->backtrack   = PStackCopy(subst->backtrack);
   
   return handle;
   
}

/*-----------------------------------------------------------------------
//
// Function: PatSymbValue()
//
//   Return the norm id assigned to f_code.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

FunCode PatSymbValue(PatternSubst_p subst, FunCode f_code)
{
   assert(f_code);
   
   if(f_code > 0)
   {
      return PDArrayElementInt(subst->fun_subst, f_code);
   }
   return PDArrayElementInt(subst->var_subst, -f_code);
}

/*-----------------------------------------------------------------------
//
// Function: PatSymbolIsBound()
//
//   Return true is f_code is either bound to a symbol or should not
//   be bound at all.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool PatSymbolIsBound(PatternSubst_p subst, FunCode f_code)
{
   return (PatSymbValue(subst, f_code) != 0);
}


/*-----------------------------------------------------------------------
//
// Function: PatternSubstBacktrack()
//
//   Backtrack a pattern-subst to a given state. Return true if the
//   state differs from the current one.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool PatternSubstBacktrack(PatternSubst_p subst, PStackPointer
			   old_state)
{
   bool    res = false;
   FunCode symbol, rep_symbol;

   assert(subst);
   assert(old_state <= PStackGetSP(subst->backtrack));
   
   while(PStackGetSP(subst->backtrack) < old_state)
   {
      symbol = PStackPopInt(subst->backtrack);
      
      if(symbol < 0)
      {
	 rep_symbol = PDArrayElementInt(subst->var_subst, -symbol);
	 assert(subst->used_vars == rep_symbol);
	 subst->used_vars++;
	 PDArrayAssignInt(subst->var_subst, -symbol, 0);
      }
      else
      {
	 int arity, count;
	 IntOrP *tmp;

	 rep_symbol = PDArrayElementInt(subst->fun_subst, -symbol);
	 arity = PatternIdGetArity(rep_symbol);
	 count = PatternIdGetIdent(rep_symbol);
	 
	 tmp = PDArrayElementRef(subst->used_idents, arity);
	 assert(tmp->i_val == count);	 
	 tmp->i_val--;
	 PDArrayAssignInt(subst->fun_subst, symbol, 0);
      }

      res = true;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PatternTermCompute()
//
//   Extend subst to make term into a pattern. Return true if a new
//   renaming has been added.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

PStackPointer PatternTermCompute(PatternSubst_p subst, Term_p term)
{
   int i;
   bool res, tmp;
   
   assert(term);
   assert(subst);

   res = extend_pat_subst(subst, term->f_code, term->arity);
   
   for(i=0; i< term->arity; i++)
   {
      assert(!TermIsVar(term));
      assert(term->args[i]);
      tmp = PatternTermCompute(subst, term->args[i]);
      res = res||tmp;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PatternTermCompare()
//
//   Compare two term patterns.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

CompareResult PatternTermCompare(PatternSubst_p subst1, Term_p t1,
				 PatternSubst_p subst2, Term_p t2)
{
   CompareResult res;

   res = pat_term_size_compare(t1, t2);
   assert(res!=to_uncomparable);
   
   if(res != to_equal)
   {
      return res;
   }
   else
   {
      PStack_p stack = PStackAlloc();
      int i;
      
      PStackPushP(stack, t1);
      PStackPushP(stack, t2);
      
      while(!PStackEmpty(stack))
      {
	 t2 = PStackPopP(stack);
	 t1 = PStackPopP(stack);
	 
	 assert(t1->arity == t2->arity);

	 res = pat_symbol_compare(subst1, t1->f_code, subst2,
				  t2->f_code);
	 if(res != to_equal)
	 {
	    break;
	 }
	 for(i=0; i<t1->arity; i++)
	 {
	    PStackPushP(stack, t1->args[i]);
	    PStackPushP(stack, t2->args[i]);
	 }
      }
      PStackFree(stack);
      return res;     
   }
}


/*-----------------------------------------------------------------------
//
// Function: PatternTermPairCompute()
//
//   Compute a pattern subst for a given equation.
//
// Global Variables: -
//
// Side Effects    : Changes subst.
//
/----------------------------------------------------------------------*/
PStackPointer PatternTermPairCompute(PatternSubst_p subst, Eqn_p eqn,
                                     PatEqnDirection direction)
{
   bool res, tmp;
   
   if(!direction)
   {
      res = PatternTermCompute(subst, eqn->lterm);
      tmp = PatternTermCompute(subst, eqn->rterm);
      res = res || tmp;
   }
   else
   {
      res = PatternTermCompute(subst, eqn->rterm);
      tmp = PatternTermCompute(subst, eqn->lterm);
      res = res || tmp;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PatternTermPairCompare()
//
//   Compare two equation patterns (described by pattern-subst, eqn,
//   direction). 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

CompareResult PatternTermPairCompare(PatternSubst_p subst1, Eqn_p
				     eqn1, PatEqnDirection dir1,
				     PatternSubst_p subst2, Eqn_p
				     eqn2, PatEqnDirection dir2)
{
   Term_p l1,l2,r1,r2;
   CompareResult res;

   res = EqnStandardWeight(eqn2) - EqnStandardWeight(eqn1);

   if(res)
   {
      return Q_TO_PART(res);
   }

   l1 = PatEqnLTerm(eqn1,dir1);
   r1 = PatEqnRTerm(eqn1,dir1);
   l2 = PatEqnLTerm(eqn2,dir2);
   r2 = PatEqnRTerm(eqn2,dir2);

   res = pat_term_size_compare(l1, l2);
   if(res!=to_equal)
   {
      return res;
   }
   res = pat_term_size_compare(r1, r2);
   if(res!=to_equal)
   {
      return res;
   }

   if(EqnIsPositive(eqn1))
   {
      if(EqnIsNegative(eqn2))
      {
	 return to_greater;
      }
   }
   else
   {
      if(EqnIsPositive(eqn2))
      {
	 return to_lesser;
      }
   }

   res = PatternTermCompare(subst1, l1, subst2, l2);
   if(res != to_equal)
   {
      return res;
   }
   res = PatternTermCompare(subst1, r1, subst2, r2);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PatternLitListCompute()
//
//   Compute a pattern-subst for the list (well, array) of literals).
//
// Global Variables: -
//
// Side Effects    : Changes subst
//
/----------------------------------------------------------------------*/

PStackPointer PatternLitListCompute(PatternSubst_p subst, PDArray_p
                                    lit_order, PDArray_p eqn_order)
{
   bool  res = false, tmp;
   Eqn_p eqn;
   int   i;

   for(i=0; (eqn = PDArrayElementP(lit_order, i)); i++)
   {
      tmp = PatternTermPairCompute(subst, eqn,
				   PDArrayElementInt(eqn_order, i));
      res = res || tmp;
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: PatternTermPrint()
//
//   Print the pattern-term to out. Supports only standard E syntax.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void PatternTermPrint(FILE* out, PatternSubst_p subst, Term_p term,
		      Sig_p sig)
{   
   FunCode id;
   int  i;

   id = PatSymbValue(subst, term->f_code);
   if(TermIsVar(term))
   {      
      if(!id)
      {
	 fprintf(out, "X%ld", -term->f_code);
      }
      else
      {
	 fprintf(out, "Xn%ld", -(id-NORM_VAR_INIT));
      }
   }
   else
   {
      if(id >= NORM_SYMBOL_LIMIT)
      {
	 fprintf(out, "f%ld_%ld",
		 PatternIdGetArity(id),
		 PatternIdGetIdent(id));	 
      }
      else
      {
	 fprintf(out, SigFindName(sig, term->f_code));
      }
      if(term->arity)
      {
	 fprintf(out, "(");
	 PatternTermPrint(out, subst, term->args[0], sig);

	 for(i=1; i<term->arity; i++)
	 {
	    fprintf(out, ",");
	    PatternTermPrint(out, subst, term->args[i], sig);
	 }
	 fprintf(out, ")");	 
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: PatternEqnPrint()
//
//   Print a pattern equation in the most reasonable form.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void PatternEqnPrint(FILE* out, PatternSubst_p subst, Eqn_p eqn,
		     PatEqnDirection direction)
{
   if(EqnIsEquLit(eqn))
   {
      if(!direction)
      {
	 PatternTermPrint(out, subst, eqn->lterm, eqn->bank->sig);
	 fputs(EqnIsPositive(eqn)?"=":"!=", out);
	 PatternTermPrint(out, subst, eqn->rterm, eqn->bank->sig);
      }
      else
      {
	 PatternTermPrint(out, subst, eqn->rterm, eqn->bank->sig);
	 fputs(EqnIsPositive(eqn)?"=":"!=", out);
	 PatternTermPrint(out, subst, eqn->lterm, eqn->bank->sig);
      }
   }
   else
   {
      if(!EqnIsPositive(eqn))
      {
	 fputc('~', out);
      }
      PatternTermPrint(out, subst, eqn->lterm, eqn->bank->sig);
   }
}


/*-----------------------------------------------------------------------
//
// Function: PatternClausePrint()
//
//   Print the clause pattern represented by lit_order, eqn_order and
//   subst and print it as a LOP list of literals .
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void PatternClausePrint(FILE* out, PatternSubst_p subst, PDArray_p
			lit_order, PDArray_p eqn_order)
{
   int i;
   Eqn_p eqn;

   eqn = PDArrayElementP(lit_order,0);
   if(eqn)
   {
      PatternEqnPrint(out, subst, eqn,
		      PDArrayElementInt(eqn_order,0));
      for(i=1; (eqn = PDArrayElementP(lit_order, i)); i++)
      {
	 fputc(',', out);
	 PatternEqnPrint(out, subst, eqn,
			 PDArrayElementInt(eqn_order,0));
      }
   }
   fputs(" <- .", out);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


