/*-----------------------------------------------------------------------

File  : classify_problem.c

Author: Stephan Schulz

Contents
 
  Read a specification and print classification and feature vector.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Sat Dec 12 22:39:18 MET 1998
    New

-----------------------------------------------------------------------*/

#include <cio_commandline.h>
#include <cio_output.h>
#include <che_clausesetfeatures.h>

/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

#define VERSION "0.2dev"

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERSION,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_DISCOUNT_VARS,
   OPT_MASK,
   OPT_ABSOLUTE,
   OPT_FEW_LIMIT,
   OPT_MANY_LIMIT,
   OPT_FEW_COUNT,
   OPT_MANY_COUNT
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},

   {OPT_VERSION,
    '\0', "version",
    NoArg, NULL,
    "Print the version number of the program."},

   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program. This differs "
    "from the output level (below) in that technical information is "
    "printed to stderr, while the output level determines which "
    "logical manipulations of the clauses are printed to stdout."},

   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},

   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    OptArg, "true",
    "Read and write PCL terms, i.e. write constants as functions of "
    "arity 0 ('a()') and treat all other identifiers as variables. "
    "otherwise use PROLOG terms (variables start with a capital "
    "letter or '_', everything else is a constant)."},

   {OPT_MASK,
    'c', "class-mask",
    ReqArg, NULL,
    "Provide a mask for the class description. A mask is a 8 letter "
    "string, with positions corresponding to the class "
    "descriptors. Any dash ('-') in the string masks out the "
    "corresponding position in the class descriptor."},

   {OPT_ABSOLUTE,
    'a', "absolute",
    OptArg, "true",
    "Use absolute values (not percentages) to determine if few, some "
    "or many clauses have a certain property."},

   {OPT_FEW_LIMIT,
    'f', "few-limit",
    ReqArg, NULL,
    "Set the limit (either an absolute integer value or a part "
    "between 0 and 1) up to which only a few clauses "
    "from a given set are considered to have a certain property."},
   
   {OPT_MANY_LIMIT,
    'm', "many-limit",
    ReqArg, NULL,
    "Set the limit (either an absolute integer value or a part "
    "between 0 and 1) from which on many clauses from "
    "a given set are considered to have a certain property."},

   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

char   *outname = NULL, 
       *mask = "aaaaa---";
double few_limit  = FEW_DEFAULT,
       many_limit = MANY_DEFAULT;
bool   absolute = false;
long   ax_some_limit     = AX_SOME_DEFAULT,
       ax_many_limit     = AX_MANY_DEFAULT, 
       lit_some_limit    = LIT_SOME_DEFAULT,
       lit_many_limit    = LIT_MANY_DEFAULT, 
       term_medium_limit = TERM_MED_DEFAULT,
       term_large_limit  = TERM_LARGE_DEFAULT;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   TB_p            terms;
   Sig_p           sig;
   ClauseSet_p     clauses;
   Scanner_p       in;    
   int             i;
   CLState_p       state;
   SpecFeatureCell features;

   assert(argv[0]);
   
   InitOutput();
   InitError("classify_problem");

   state = process_options(argc, argv);

   if(absolute)
   {
      if(few_limit < 0)
      {
	 Error("Option -f (--few-limit) requires value >=0.",
	       USAGE_ERROR);
      }
      if(many_limit < 0)
      {
	 Error("Option -m (--many-limit) requires value >=0.",
	       USAGE_ERROR);
      }
      if(few_limit == FEW_DEFAULT)
      {
	 few_limit = FEW_ABSDEFAULT;
      }
      if(many_limit == MANY_DEFAULT)
      {
	 many_limit = MANY_ABSDEFAULT;
      }
   }
   else
   {
      if((few_limit < 0.0)||(few_limit > 1.0))
      {
	 Error("Option -f (--few-limit) requires value between "
	       "0 and 1.", USAGE_ERROR);
      }
      if((many_limit < 0.0)||(many_limit > 1.0))
      {
	 Error("Option -M (--many-limit) requires value between "
	       "0 and 1.", USAGE_ERROR);
      }
   }
   if(few_limit>many_limit)
   {
      Warning("Limit for 'many' was smaller than limit for "
	      "'few'. Was set to 'few'-limit.");
   }
      
   GlobalOut = OutOpen(outname);

   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }
   
   for(i=0; state->argv[i]; i++)
   {
      sig     = SigAlloc();
      terms   = TBAlloc(TPIgnoreProps, sig);
      clauses = ClauseSetAlloc();
      in = CreateScanner(StreamTypeFile, state->argv[i] , true, NULL);
      
      ClauseSetParseList(in, clauses, terms);
      SpecFeaturesCompute(&features, clauses);
      SpecFeaturesAddEval(&features, few_limit, many_limit, absolute,
			  ax_some_limit, ax_many_limit,
			  lit_some_limit, lit_many_limit,
			  term_medium_limit, term_large_limit);

      fprintf(GlobalOut, "%s : ", state->argv[i]);
      SpecFeaturesPrint(GlobalOut, &features);
      fprintf(GlobalOut, " : ");
      SpecTypePrint(GlobalOut, &features, mask);
      fprintf(GlobalOut, "\n");
      
      DestroyScanner(in);
      ClauseSetFree(clauses);
      terms->sig = NULL;
      TBFree(terms);
      SigFree(sig);
   }
   
   CLStateFree(state);
   
   fflush(GlobalOut);
   OutClose(GlobalOut);
   
#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif
   
   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
	    break;
      case OPT_VERSION:
	    printf("classify_problem " VERSION "\n");
	    exit(NO_ERROR);
	    break;
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_MASK:
	    mask = arg;
	    if(strlen(mask)!=8)
	    {
	       Error("Option -c (--class-mask) requires 8-letter "
		     "string as an argument", USAGE_ERROR);
	    }
	    break;
      case OPT_ABSOLUTE:
	    absolute = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FEW_LIMIT:
	    few_limit = CLStateGetFloatArg(handle, arg);
	    break;
      case OPT_MANY_LIMIT:
	    many_limit = CLStateGetFloatArg(handle, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}

void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
classify_problem " VERSION "\n\
\n\
Usage: classify_problem [options] [files]\n\
\n\
Read sets of clauses and classify them according to predefined criteria.\n\
\n");
   PrintOptions(stdout, opts);
   fprintf(out, "\n\
Copyright 1998 by Stephan Schulz, schulz@informatik.tu-muenchen.de\n\
\n\
This program is a part of the support structure for the E equational\n\
theorem prover. You can find the latest version of the E distribution\n\
as well as additional information at\n\
http://wwwjessen.informatik.tu-muenchen.de/~schulz/WORK/eprover.html.\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (it should be contained in the top level\n\
directory of the distribution in the file COPYING); if not, write to\n\
the Free Software Foundation, Inc., 59 Temple Place, Suite 330,\n\
Boston, MA  02111-1307 USA\n\
\n\
The original copyright holder can be contacted as\n\
\n\
Stephan Schulz\n\
Technische Universitaet Muenchen\n\
Fakultaet fuer Informatik\n\
Arcisstrasse 20\n\
D-80290 Muenchen\n\
Germany\n\
");

}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


