/*-----------------------------------------------------------------------

File  : eprover.c

Author: Stephan Schulz

Contents
 
  Main program for the E equational theorem prover.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Tue Jun  9 01:32:15 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include <sys/time.h>
#include <unistd.h>

#include <cio_commandline.h>
#include <cio_output.h>
#include <cco_proofproc.h>
#include <cco_signals.h>

/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

#define VERSION "0.32-dev"
#define NAME ""

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERSION,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_SATURATED,
   OPT_PRINT_SAT_INFO,
   OPT_FILTER_SATURATED,
   OPT_PRINT_PID,
   OPT_SILENT,
   OPT_OUTPUTLEVEL,
   OPT_MEM_LIMIT,
   OPT_CPU_LIMIT,
   OPT_SOFTCPU_LIMIT,
   OPT_RUSAGE_INFO,
   OPT_STEP_LIMIT,
   OPT_PROC_LIMIT,
   OPT_UNPROC_LIMIT,
   OPT_TOTAL_LIMIT,
   OPT_DISCOUNT_VARS,
   OPT_NO_INFIX,
   OPT_FULL_EQ_REP,
   OPT_TPTP_PARSE,
   OPT_TPTP_PRINT,
   OPT_TPTP_FORMAT,
   OPT_NO_NEG_PARAMOD,
   OPT_NO_NONEQ_PARAMOD,
   OPT_PARAMOD_STRATEGY,
   OPT_NO_GENERATION,
   OPT_LITERAL_SELECT,
   OPT_POS_LITSEL_MIN,
   OPT_POS_LITSEL_MAX,
   OPT_NEG_LITSEL_MIN,
   OPT_NEG_LITSEL_MAX,
   OPT_ALL_LITSEL_MIN,
   OPT_ALL_LITSEL_MAX,
   OPT_PREFER_INITIAL_CLAUSES,
   OPT_FILTER_LIMIT,
   OPT_FILTER_COPIES_LIMIT,
   OPT_REWEIGHT_LIMIT,
   OPT_DELETE_BAD_LIMIT,
   OPT_ORDERING,
   OPT_TO_WEIGHTGEN,
   OPT_TO_PRECGEN,
   OPT_TO_CONSTWEIGHT,
   OPT_INDEXING,   
   OPT_RULES_GENERAL,
   OPT_FORWARD_DEMOD,
   OPT_STRONGSUBSUMPTION,
   OPT_DEFINE_WFUN,
   OPT_DEFINE_HEURISTIC,
   OPT_HEURISTIC
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},

   {OPT_VERSION,
    'V', "version",
    NoArg, NULL,
    "Print the version number of the prover."},

   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program. This differs "
    "from the output level (below) in that technical information is "
    "printed to stderr, while the output level determines which "
    "logical manipulations of the clauses are printed to stdout."},

   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},

   {OPT_SILENT,
    's', "silent",
    NoArg, NULL,
    "Equivalent to --output-level=0."},

   {OPT_OUTPUTLEVEL,
    'l', "output-level",
    ReqArg, NULL,
    "Select an output level, greater values imply more verbose "
    "output. Level 0 produces "
    "nearly no output, level 1 will output each clause as it is "
    "processed, level 2 will output generating inferences, "
    "level 3 will give a full protocol including rewrite steps and "
    "level 4 will include some internal clause renamings. Levels >= 2"
    " also imply a certain clause format (LOP with infix equality "
    "predicate) that is expected by the postprocessing tools."},
   
   {OPT_PRINT_SATURATED,
    'S', "print-saturated",
    OptArg, DEFAULT_OUTPUT_DESCRIPTOR,
    "Print the (semi-) saturated clause sets after terminating the "
    "saturation process. The argument given descrbes which parts "
    "should be printed in which order. Legal caracters are 'eigEIGaA',"
    " standing for processed positive units, processed negative units, "
    "processed non-units, unprocessed positive units, unprocessed "
    "negative units, unprocessed non-units, and two types of "
    "additional equality axioms, respectively. Equality axioms will "
    "only be printed if the original specification contained real "
    "equality. In this case, 'a' requests axioms in which a separate" 
    " substitutivity axiom is given for each argument position of a"
    " subction or predicate symbol, while 'A' requests a single "
    "substitutivity axiom (covering all positions) for each symbol."},  

   {OPT_PRINT_SAT_INFO,
    '\0', "print-sat-info",
    NoArg, NULL,
    "Print additional information (clause number, weight, etc) as a"
    "comment for clauses from the semi-saturated end system,"},

   {OPT_FILTER_SATURATED,
    '\0', "filter-saturated",
    OptArg, DEFAULT_FILTER_DESCRIPTOR,
    "Filter the  (semi-) saturated clause sets after terminating the "
    "saturation process. The argument is a string describing which "
    "operations to take (and in which order). Options are 'u' (remove"
    " all clauses with more than one literal), 'c' (delete all but one"
    " copy of identical clauses, 'n', 'r', 'f' (forward contraction, "
    "unit-subsumption only, no rewriting, rewriting with rules only,"
    " full rewriting, respectively), and 'N', 'R' and 'F' (as their"
    " lower case counterparts, but with non-unit-subsumption enabled"
    " as well)."}, 
   
   {OPT_PRINT_PID,
    '\0', "print-pid",
    NoArg, NULL,
    "Print the process id of the prover as a comment after option "
    "processing."},

   {OPT_MEM_LIMIT,
    'm', "memory-limit",
    OptArg, "120",
    "Limit the memory the prover may use. The optional argument is "
    "the allowed amount of memory in MB. This option may not work "
    "everywhere, due to broken and/or strange behaviour of setrlimit() "
    "in some UNIX implementations. It does work under all tested "
    "versions of Solaris and GNU/Linux."},   

   {OPT_CPU_LIMIT,
    '\0', "cpu-limit",
    OptArg, "300",
    "Limit the cpu time the prover should run. The optional argument "
    "is the CPU time in seconds. The prover will terminate immediately"
    " after reaching the time limit, regardless of inernal state. This"
    " option may not work " 
    "everywhere, due to broken and/or strange behaviour of setrlimit() "
    "in some UNIX implementations. It does work under all tested "
    "versions of Solaris, HP-UX and GNU/Linux. As a side effect, this "
    "option will inhibit core file writing."}, 

   {OPT_SOFTCPU_LIMIT,
    '\0', "soft-cpu-limit",
    OptArg, "300",
    "Limit the cpu time the prover should spend in the main saturation"
    "phase. The prover will then terminate gracefully, i.e. it will "
    "perform post-processing, " 
    "filtering and printing of unprocessed clauses, if these options"
    "are selected. Note"
    " that for some filtering options (in particular those which "
    "perform full subsumption), the postprocessing time may well be"
    "larger than the saturation time. This option is particularly "
    "useful if you want to use E as a preprocessor or lemma generator "
    "in a larger system."},  

   {OPT_RUSAGE_INFO,
    'R', "resources-info",
    NoArg, NULL,
    "Give some information about the resources used by the prover. "
    "You will usually get CPU time information. On systems returning "
    "more information with the rusage() system call, you will also "
    "get information about memory consumption."},

   {OPT_STEP_LIMIT,
    'C', "processed-clauses-limit",
    ReqArg, NULL,
    "Set the maximal number of clauses to process."},
   
   {OPT_PROC_LIMIT,
    'P', "processed-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of processed clauses."},

   {OPT_UNPROC_LIMIT,
    'U', "unprocessed-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of unprocessed clauses."},
   
   {OPT_TOTAL_LIMIT,
    'T', "total-clause-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of all clauses."},

   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    NoArg, NULL,
    "Read and write PCL terms, i.e. write constants as functions of "
    "arity 0 ('a()') and treat all other identifiers as variables. "
    "The default is to use PROLOG terms (variables start with a "
    "capital letter or '_', everything else is a constant)."},

   {OPT_NO_INFIX,
    'n', "eqn-no-infix",
    NoArg, NULL,
    "Print equations in prefix notation equal(x,y)."},

   {OPT_FULL_EQ_REP,
    'e', "full-equational-rep",
    NoArg, NULL,
    "Print all literals as equations, even non-equational ones."},

   {OPT_TPTP_PARSE,
    '\0', "tptp-in",
    NoArg, NULL,
    "Parse TPTP format instead of lop."},

   {OPT_TPTP_PRINT,
    '\0', "tptp-out",
    NoArg, NULL,
    "Print TPTP format instead of lop. Implies --eqn-no-infix and "
    "will ignore --full-equational-rep."},

   {OPT_TPTP_FORMAT,
    '\0', "tptp-format",
    NoArg, NULL,
    "Equivalent to --tptp-in and --tptp-out."},

   {OPT_NO_NEG_PARAMOD,
    'N', "no-paramod-into-negatives",
    NoArg, NULL,
    "Refrain from paramodulating into negative literals."},

   {OPT_NO_NONEQ_PARAMOD,
    'E', "paramod-equational-literals-only",
    NoArg, NULL,
    "Don't paramodulate from or into non-equational literals."},

   {OPT_NO_GENERATION,
    '\0', "no-generation",
    NoArg, NULL,
    "Don't perform any generating inferences (equivalent to "
    "--literal-selection-strategy=NoGeneration)."},

   {OPT_PARAMOD_STRATEGY,
    '\0', "paramod-strategy",
    OptArg, "Normal",
    "This option has been obsoleted and is mapped to literal"
    " selection strategies (see the next option). Determine the "
    "paramod-strategy: Normal (paramodulate from all " 
    "positive literals into all literals), WithUnitsOnly "
    "(paramodulate with positive units into all literals), "
    "PosUnitStrategy (paramodulate with units only, into positive "
    "unit clauses and negative literals in non-unit-clauses) or "
    "NoGeneration (don't perform any generating inferences). The two "
    "unit strategies are only complete for Horn problems, but should "
    "be fairly efficient for this case. If one of these is selected, "
    "the literal ordering is also changed: Negative and positive"
    "literals are made uncomparable. The last strategy will generate an"
    " interreduced clause set with all subsumed clauses removed, but will"
    " of course not be complete for any interesting class of problems."},	

   {OPT_LITERAL_SELECT,
    'W', "literal-selection-strategy",
    ReqArg, NULL,
    "Choose a strategy for selection of negative literals. There are "
    "two special values for this option: NoSelection will select no "
    "literal (i.e. perform normal superposition) and "
    "NoGeneration will inhibit all generatin inferences. For a list "
    "of the other (hopefully self-documenting) values run "
    "'eprover -W none'. There are two variants of each strategy. The "
    "one prefixed with 'P' will allow paramodulations into maximal "
    "positive literals in addition to paramodulations into maximal "
    "selected negative literals"},	

   {OPT_POS_LITSEL_MIN,
    '\0', "selection-pos-min",
    ReqArg, NULL,
    "Set a lower limit for the number of positive literals a clause must"
    " have to be electable for literal selection."},

   {OPT_POS_LITSEL_MAX,
    '\0', "selection-pos-max",
    ReqArg, NULL,
    "Set a upper limit for the number of positive literals a clause can"
    " have to be electable for literal selection."},

   {OPT_NEG_LITSEL_MIN,
    '\0', "selection-neg-min",
    ReqArg, NULL,
    "Set a lower limit for the number of negative literals a clause must"
    " have to be electable for literal selection."},

   {OPT_NEG_LITSEL_MAX,
    '\0', "selection-neg-max",
    ReqArg, NULL,
    "Set a upper limit for the number of negative literals a clause can"
    " have to be electable for literal selection."},

   {OPT_ALL_LITSEL_MIN,
    '\0', "selection-all-min",
    ReqArg, NULL,
    "Set a lower limit for the number of literals a clause must"
    " have to be electable for literal selection."},

   {OPT_ALL_LITSEL_MAX,
    '\0', "selection-all-max",
    ReqArg, NULL,
    "Set an upper limit for the number of literals a clause must"
    " have to be electable for literal selection."},

   {OPT_PREFER_INITIAL_CLAUSES,
    '\0', "prefer-initial-clauses",
    NoArg, NULL,
    "Always process all initial clauses first."},

   {OPT_HEURISTIC,
    'x', "expert-heuristic",
    ReqArg, NULL,
    "Select one of the heuristics. Currently available: Weight, "
    "RWeight, FIFO, LIFO (will make prover incompete!), Default "
    "(will be redefinded if option -H is used to define anonymous "
    "heuristic)."},

   {OPT_FILTER_LIMIT,
    '\0', "filter-limit",
    OptArg, "1000000",
    "Set the number of 'storage units' in new unprocessed clauses"
    " after which the set of" 
    " unprocessed clauses will be filtered against the processed"
    " clauses to eleminate redundant clauses. A clause with n"
    " literals occupies n+3 storage units, a term cell occupies 3"
    " storage units. As a rule, about 1 million storage units fit into"
    " 64 MB of memory, and the prover has a base footprint of about"
    " 1-2 MB, depending on architeture and other details. "},

   {OPT_FILTER_COPIES_LIMIT,
    '\0', "filter-copies-limit",
    OptArg, "800000",
    "Set the number of storage units in new unprocessed clauses after"
    " which the set of unprocessed clauses will be filtered for"
    " equivalent copies of clauses (see above). As this operation is"
    " cheaper, you may want to set this limit lower than"
    " --filter-limit."},

   {OPT_DELETE_BAD_LIMIT,
    '\0', "delete-bad-limit",
    OptArg, "1500000",
    "Set the number of storage units after which bad clauses are"
    " deleted without further consideration. This causes the prover to"
    " be potentially incomplete, but will allow you to limit the"
    " maximum amount of memory used fairly well. The prover will tell"
    " you if a proof attempt failed due to the incompletness"
    " introduced by this option. It is recommended to set this limit"
    " significantly higher than --filter-limit or"
    " --filter-copies-limit."}, 

   {OPT_REWEIGHT_LIMIT,
    '\0', "reweight-limit",
    OptArg, "30000",
    "Set the number of new unprocessed clauses after which the set of"
    " unprocessed clauses will reevaluated."},

   {OPT_ORDERING,
    't', "term-ordering",
    ReqArg, NULL,
    "Select an ordering type (currently Auto, Optimize, LPO, KBO or "
    "KBO1)."},

   {OPT_TO_WEIGHTGEN,
    'w', "order-weight-generation",
    ReqArg, NULL,
    "Select a method for the generation of weights for use with the "
    "term ordering. Run 'eprover -w none' for a list of options."},

   {OPT_TO_PRECGEN,
    'G', "order-precedence-generation",
    ReqArg, NULL,
    "Select a method for the generation of a precedence for use with "
    "the term ordering. Run 'eprover -G none' for a list of "
    "options."}, 

   {OPT_TO_CONSTWEIGHT,
    'c', "order-constant-weight", 
    ReqArg, NULL,
    "Set a special weight > 0 for constants in the term ordering."
    "By default, constants are treated like other function symbols."
   },
   
   {OPT_INDEXING,
    'p', "no-pdt-indexing",
    NoArg, NULL,
    "Use conventional matching for forward rewriting and subsumption "
    "(default is to use perfect discrimination trees)."}, 

   {OPT_RULES_GENERAL,
    'g', "prefer-general-demodulators",
    NoArg, NULL,
    "If using perfect discrimination tree indexing, prefer general "
    "demodulators. By default, use specialized demodulators."},
   
   {OPT_FORWARD_DEMOD,
    'F', "forward_demod_level",
    ReqArg, NULL,
    "Set the desired level rewriting of unprocessed clauses. A value "
    "of 0 means no rewriting, 1 indicates to use rules (orientable "
    "equations) only, 2 indicates full rewriting with rules and "
    "instances of unorientable equations. Default is 2."},

   {OPT_STRONGSUBSUMPTION,
    'u', "strong-forward-subsumption",
    NoArg, NULL,
    "Try multiple positions and unit-equations to try to subsume a "
    "single new clause. Default is to search for a single position."},

   {OPT_DEFINE_WFUN,
    'D', "define-weight-function",
    ReqArg, NULL,
    "Define weight functions. Later definitions override previous definitions."},

   {OPT_DEFINE_HEURISTIC,
    'H', "define-heuristic",
    ReqArg, NULL,
    "Define a heuristic. Later definitions override previous definitions."},

   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

char              *outname = NULL;
HeuristicParms_p  h_parms;
bool              print_sat = false,
                  filter_sat = false,
                  print_rusage = false,
                  print_pid = false,
                  outinfo = false;
long              step_limit = LONG_MAX, 
                  proc_limit = LONG_MAX,
                  unproc_limit = LONG_MAX, 
                  total_limit = LONG_MAX;
char              *outdesc = DEFAULT_OUTPUT_DESCRIPTOR,
                  *filterdesc = DEFAULT_FILTER_DESCRIPTOR;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p        state;
   Scanner_p        in;    
   ProofState_p     proofstate;
   ProofControl_p   proofcontrol;
   int              i;
   Clause_p         success, filter_success;
   
   assert(argv[0]);

   InitOutput();
   InitError("eprover");
   
   if (signal(SIGXCPU, ESignalHandler) == SIG_ERR)
   {
      TmpErrno = errno;
      SysError("Unable to set up signal handler", SYS_ERROR);
   }
   
   h_parms = HeuristicParmsAlloc();

   state = process_options(argc, argv);

   GlobalOut = OutOpen(outname);

   if(print_pid)
   {
      fprintf(GlobalOut, "# Pid: %d\n", (int)getpid());
      fflush(GlobalOut);
   }

   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }

   proofstate = ProofStateAlloc();
   proofcontrol = ProofControlAlloc();
   
   for(i=0; state->argv[i]; i++)
   {
      in = CreateScanner(StreamTypeFile, state->argv[i] , true, NULL);
      ClauseSetParseList(in, proofstate->axioms,
			 proofstate->original_terms);
      DestroyScanner(in); 
   }
   VERBOUT2("Specification read\n");
   ProofControlInit(proofstate, proofcontrol, h_parms);
   ProofStateInit(proofstate, proofcontrol, h_parms);
   DemodOrderCriterion = GetHeuristic("LIFO", proofstate, proofcontrol, h_parms);

   VERBOUT2("Prover state initialized\n");
   
   success = Saturate(proofstate, proofcontrol, step_limit,
		      proc_limit, unproc_limit, total_limit);
   
   if(filter_sat)
   {
      filter_success = ProofStateFilterUnprocessed(proofstate,
						   proofcontrol,
						   filterdesc);
      if(filter_success)
      {
	 success = filter_success;
      }
   }
   
   if(success)
   {
      ClauseDefaultDocumentProcessing(success, CP_PROOF);
      fprintf(GlobalOut, "\n# Proof found!\n");
   }
   else
   {
      ClauseSetDocumentProcessing(GlobalOut, OutputLevel,
				  proofstate->processed_pos_rules,
				  CP_FINAL); 
      ClauseSetDocumentProcessing(GlobalOut, OutputLevel,
				  proofstate->processed_pos_eqns,
				  CP_FINAL); 
      ClauseSetDocumentProcessing(GlobalOut, OutputLevel,
				  proofstate->processed_neg_units,
				  CP_FINAL); 
      ClauseSetDocumentProcessing(GlobalOut, OutputLevel,
				  proofstate->processed_non_units,
				  CP_FINAL);
      if(ClauseSetEmpty(proofstate->unprocessed))
      {
	 if(!ParamodOverlapIntoNegativeLiterals||
	    !ParamodOverlapNonEqLiterals||
	    (proofcontrol->selection_strategy==SelectNoGeneration))
	 {
	    fprintf(GlobalOut, 
		    "\n# Clause set closed under "
		    "restricted calculus!\n");
	 }
	 else if(proofstate->state_is_complete)
	 {
	    fprintf(GlobalOut, "\n# No proof found!\n");
	 }
	 else
	 {
	    fprintf(GlobalOut, "\n# Failure: Out of unprocessed clauses!\n");	    
	 }
      }
      else
      {
	 fprintf(GlobalOut, "\n# Failure: User resource limit exceeded!\n");
      }
   }
   if(print_sat)
   {
      if(success)
      {
	 fprintf(GlobalOut, "# Saturated system contains the empty clause:\n");
	 ClausePrint(GlobalOut, success, true);
	 fputc('\n',GlobalOut);	 
	 fputc('\n',GlobalOut);	 
      }
      ProofStatePrintSelective(GlobalOut, proofstate, outdesc,
			       outinfo);
      fprintf(GlobalOut, "\n");
   }	 
   
   if(success)
   {
      ClauseFree(success);
   }
   fflush(GlobalOut);
   
   if(OutputLevel)
   {
      ProofStateStatisticsPrint(GlobalOut, proofstate);
   }
#ifndef FAST_EXIT
   ProofControlFree(proofcontrol);
   ProofStateFree(proofstate);
   CLStateFree(state);
   HeuristicParmsFree(h_parms);
#endif

   if(print_rusage)
   {
      PrintRusage(GlobalOut);
   }
   fflush(GlobalOut);
   OutClose(GlobalOut);
   
#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif
   
   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   struct rlimit     limit = {RLIM_INFINITY, RLIM_INFINITY};
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
	    break;
      case OPT_VERSION:
	    printf("E " VERSION "\n");
	    exit(NO_ERROR);
	    break;
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_SILENT:
	    OutputLevel = 0;
	    break;
      case OPT_OUTPUTLEVEL:
	    OutputLevel = CLStateGetIntArg(handle, arg);
	    if(OutputLevel >= 2)
	    {
	       TermPrologVars = true;
	       EqnUseInfix = true;
	       EqnFullEquationalRep = false;
	       TPTPFormatPrint = false;
	    }
	    break;
      case OPT_PRINT_SATURATED:
	    outdesc = arg;
	    if(!CheckLetterString(outdesc, "eigEIGaA"))
	    {
	       Error("Illegal argument to option -S"
		     " (--print-saturated)", 
		     USAGE_ERROR);
	    }
	    print_sat = true;
	    break;	    
      case OPT_PRINT_SAT_INFO:
	    outinfo = true;
	    break;
      case OPT_FILTER_SATURATED:
	    filterdesc = arg;
	    if(!CheckLetterString(filterdesc, "ucnrfNRF"))
	    {
	       Error("Illegal argument to option --filter-saturated",
		     USAGE_ERROR);
	    }
	    filter_sat = true;	    
	    break;
      case OPT_PRINT_PID:
	    print_pid = true;
	    break;
      case OPT_MEM_LIMIT:
	    h_parms->mem_limit = CLStateGetIntArg(handle, arg);
	    limit.rlim_cur = MEGA*h_parms->mem_limit;
	    if(setrlimit(RLIMIT_DATA, &limit))
	    {
	       TmpErrno = errno;
	       SysError("Unable to set memory limit", SYS_ERROR);
	    }
	    break;
      case OPT_CPU_LIMIT:
      case OPT_SOFTCPU_LIMIT:
	    limit.rlim_cur = CLStateGetIntArg(handle, arg);
	    if(setrlimit(RLIMIT_CPU, &limit))
	    {
	       TmpErrno = errno;
	       SysError("Unable to set cpu time limit", SYS_ERROR);
	    }
	    limit.rlim_cur = 0;
	    if(setrlimit(RLIMIT_CORE, &limit))
	    {
	       TmpErrno = errno;
	       SysError("Unable to prevent core dumps", SYS_ERROR);
	    }
	    SoftTimeLimit = (handle->option_code == OPT_SOFTCPU_LIMIT);
	    break;
      case OPT_RUSAGE_INFO:
	    print_rusage = true;
	    break;
      case OPT_STEP_LIMIT:
	    step_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PROC_LIMIT:
	    proc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_UNPROC_LIMIT:
	    unproc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_TOTAL_LIMIT:
	    total_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = false;
	    break;
      case OPT_NO_INFIX:
	    EqnUseInfix = false;
	    break;
      case OPT_FULL_EQ_REP:
	    EqnFullEquationalRep = true;
	    break;
      case OPT_TPTP_PARSE:
	    TPTPFormatParse = true;
	    break;
      case OPT_TPTP_PRINT:
	    TPTPFormatPrint = true;
	    EqnFullEquationalRep = false;
	    EqnUseInfix = false;
	    break;
      case OPT_TPTP_FORMAT:
	    TPTPFormatParse = true;
	    TPTPFormatPrint = true;
	    EqnFullEquationalRep = false;
	    EqnUseInfix = false;
	    break;	    
      case OPT_NO_NEG_PARAMOD:
	    ParamodOverlapIntoNegativeLiterals = false;
	    break;	    
      case OPT_NO_NONEQ_PARAMOD:
	    ParamodOverlapNonEqLiterals = false;
	    break;
      case OPT_NO_GENERATION:
	    h_parms->selection_strategy=SelectNoGeneration;
	    break;
      case OPT_PARAMOD_STRATEGY:
	    if(strcmp(arg, "Normal")==0)
	    {
	       h_parms->selection_strategy = SelectNoLiterals;
	    }
	    else if(strcmp(arg, "WithUnitsOnly")==0)
	    {
	       h_parms->selection_strategy = SelectPosNegativeLiterals;
	    }
	    else if(strcmp(arg, "PosUnitStrategy")==0)
	    {
	       h_parms->selection_strategy = SelectNegativeLiterals;
	    }
	    else if(strcmp(arg, "NoGeneration")==0)
	    {
	       h_parms->selection_strategy = SelectNoGeneration;
	    }
	    else
	    {
	       Error("Option --paramod-strategy) requires Normal, "
		     "WithUnitsOnly, PosUnitStrategy or NoGeneration as an "
		     "argument", 
		     USAGE_ERROR);
	    }
	    break;	    
      case OPT_LITERAL_SELECT:
	    h_parms->selection_strategy = GetLitSelFun(arg);
	    if(!h_parms->selection_strategy)
	    {
	       DStr_p err = DStrAlloc();
	       DStrAppendStr(err, 
			     "Wrong argument to option -W "
			     "(--literal-selection-strategy). Possible "
			     "values: "); 
	       DStrAppendStrArray(err, LiteralSelectionFunNames, ", ");
	       Error(DStrView(err), USAGE_ERROR);
	       DStrFree(err);
	    }		      
	    break;
      case OPT_POS_LITSEL_MIN:
	    h_parms->pos_lit_sel_min = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_POS_LITSEL_MAX:
	    h_parms->pos_lit_sel_max = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_NEG_LITSEL_MIN:
	    h_parms->neg_lit_sel_min = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_NEG_LITSEL_MAX:
	    h_parms->neg_lit_sel_max = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_ALL_LITSEL_MIN:
	    h_parms->all_lit_sel_min = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_ALL_LITSEL_MAX:
	    h_parms->all_lit_sel_max = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PREFER_INITIAL_CLAUSES:
	    h_parms->prefer_initial_clauses = true;
	    break;
      case OPT_HEURISTIC:
	    h_parms->heuristic_name = arg;
	    break;
      case OPT_FILTER_LIMIT:
	    h_parms->filter_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_FILTER_COPIES_LIMIT:
	    h_parms->filter_copies_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_DELETE_BAD_LIMIT:
	    h_parms->delete_bad_limit = CLStateGetIntArg(handle, arg);
	    break;
         case OPT_REWEIGHT_LIMIT:
	    h_parms->reweight_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_ORDERING:	    
	    if(strcmp(arg, "Auto")==0)
	    {
	       h_parms->ordertype = AUTO;
	    }
	    else if(strcmp(arg, "Optimize")==0)
	    {
	       h_parms->ordertype = OPTIMIZE_AX;
	    }
	    else if(strcmp(arg, "LPO")==0)
	    {
	       h_parms->ordertype = LPO;
	    }
	    else if(strcmp(arg, "KBO1")==0)
	    {
	       h_parms->ordertype = KBO;
	    }
	    else if(strcmp(arg, "KBO")==0)
	    {
	       h_parms->ordertype = KBO2;
	    }
	    else
	    {
	       Error("Option -t (--term-ordering) requires Auto, "
		     "Optimize,LPO, KBO or KBO1 as an argument",
		     USAGE_ERROR);
	    }
	    break;
      case OPT_TO_WEIGHTGEN:
	    h_parms->to_weight_gen = TOTranslateWeightGenMethod(arg);
	    if(!h_parms->to_weight_gen)
	    {
	       DStr_p err = DStrAlloc();
	       DStrAppendStr(err, 
			     "Wrong argument to option -w "
			     "(--order-weight-generation). Possible "
			     "values: "); 
	       DStrAppendStrArray(err, TOWeightGenNames, ", ");
	       Error(DStrView(err), USAGE_ERROR);
	       DStrFree(err);
	    }		      
	    break;
      case OPT_TO_PRECGEN:
	    h_parms->to_prec_gen = TOTranslatePrecGenMethod(arg);
	    if(!h_parms->to_prec_gen)
	    {
	       DStr_p err = DStrAlloc();
	       DStrAppendStr(err, 
			     "Wrong argument to option -G "
			     "(--order-precedence-generation). Possible "
			     "values: "); 
	       DStrAppendStrArray(err, TOPrecGenNames, ", ");
	       Error(DStrView(err), USAGE_ERROR);
	       DStrFree(err);
	    }		      
	    break;
      case OPT_TO_CONSTWEIGHT:
	    h_parms->to_const_weight = CLStateGetIntArg(handle, arg); 
	    if(h_parms->to_const_weight<=0)
	    {
	       Error("Argument to option -c (--order-constant-weight) "
		     "has to be > 0", USAGE_ERROR);
	    }
	    break;
      case OPT_INDEXING:
	    h_parms->use_indexing = false;
	    break;
      case OPT_RULES_GENERAL:
	    PDTreePreferGeneral = true;
	    break;
      case OPT_FORWARD_DEMOD:
	    h_parms->forward_demod = CLStateGetIntArg(handle, arg);
	    if((h_parms->forward_demod > 2) || 
	       (h_parms->forward_demod < 0))
	    {
	       Error("Option -F (--forward_demod_level) requires "
		     "argument from {0..2}", USAGE_ERROR);
	    }
	    break;	       
      case OPT_STRONGSUBSUMPTION:
	    StrongUnitForwardSubsumption = true;
	    break;
      case OPT_DEFINE_WFUN:
	    PStackPushP(h_parms->wfcb_definitions, arg);
	    break;
      case OPT_DEFINE_HEURISTIC:
	    PStackPushP(h_parms->hcb_definitions, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}

void print_help(FILE* out)
{
   fprintf(out, "\n\
E " VERSION " \"" NAME "\"\n\
\n\
Usage: eprover [options] [files]\n\
\n\
Read a set of clauses and try to refute it.\n\
\n");
   PrintOptions(stdout, opts);
   fprintf(out, "\n\
Copyright 1998, 1999 by Stephan Schulz, schulz@informatik.tu-muenchen.de\n\
\n\
You can find the latest version of E and additional information at\n\
http://wwwjessen.informatik.tu-muenchen.de/~schulz/WORK/eprover.html.\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (it should be contained in the top level\n\
directory of the distribution in the file COPYING); if not, write to\n\
the Free Software Foundation, Inc., 59 Temple Place, Suite 330,\n\
Boston, MA  02111-1307 USA\n\
\n\
The original copyright holder can be contacted as\n\
\n\
Stephan Schulz\n\
Technische Universitaet Muenchen\n\
Fakultaet fuer Informatik\n\
D-80290 Muenchen\n\
Germany\n\
");

}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


